@@ -1,4 +1,3 @@
-use std::borrow::Cow;
 use std::collections::{BTreeMap, HashMap};
 use std::ffi::OsString;
 use std::fs::{self, File};
@@ -6,6 +5,7 @@ use std::io::{self, BufWriter, Cursor, Read, Write};
 use std::path::{Path, PathBuf};
 use std::time::Duration;
 
+use once_cell::unsync::OnceCell;
 use yansi::Color::{Green, Red};
 use yansi::Paint;
 use zip::ZipArchive;
@@ -14,20 +14,24 @@ use crate::config::Config;
 use crate::error::{Error, Result};
 use crate::util::{self, info_end, info_start, infoln, warnln, Dedup};
 
-const BASE_ARCHIVE_URL: &str =
-    "https://raw.githubusercontent.com/tldr-pages/tldr-pages.github.io/main/assets";
-const CHECKSUMS: &str =
-    "https://raw.githubusercontent.com/tldr-pages/tldr-pages.github.io/main/assets/tldr.sha256sums";
+pub const ENGLISH_DIR: &str = "pages.en";
 const USER_AGENT: &str = concat!(env!("CARGO_PKG_NAME"), '/', env!("CARGO_PKG_VERSION"));
-const ENGLISH_DIR: &str = "pages.en";
 
 type PagesArchive = ZipArchive<Cursor<Vec<u8>>>;
 
-pub struct Cache<'a>(&'a Path);
+pub struct Cache<'a> {
+    dir: &'a Path,
+    platforms: OnceCell<Vec<OsString>>,
+    age: OnceCell<Duration>,
+}
 
 impl<'a> Cache<'a> {
     pub fn new(dir: &'a Path) -> Self {
-        Self(dir)
+        Self {
+            dir,
+            platforms: OnceCell::new(),
+            age: OnceCell::new(),
+        }
     }
 
     /// Get the default path to the cache.
@@ -35,37 +39,70 @@ impl<'a> Cache<'a> {
         dirs::cache_dir().unwrap().join(env!("CARGO_PKG_NAME"))
     }
 
-    /// Return `true` if the English pages directory exists.
-    pub fn english_dir_exists(&self) -> bool {
-        self.0.join(ENGLISH_DIR).is_dir()
+    /// Return `true` if the specified subdirectory exists in the cache.
+    pub fn subdir_exists(&self, sd: &str) -> bool {
+        self.dir.join(sd).is_dir()
     }
 
-    /// Download the tldr pages archives.
-    fn download_and_verify(languages: &[String]) -> Result<BTreeMap<String, PagesArchive>> {
-        let agent = ureq::builder().user_agent(USER_AGENT).build();
+    /// Send a GET request with the provided agent and return the response body.
+    fn get_asset(agent: &ureq::Agent, url: &str) -> Result<Vec<u8>> {
+        info_start!("downloading '{}'... ", url.split('/').last().unwrap());
+        let resp = agent.get(url).call()?;
+
+        let content_length = resp
+            .header("Content-Length")
+            .unwrap_or_default()
+            .parse()
+            .unwrap_or_default();
+        let mut buf = Vec::with_capacity(content_length);
+        resp.into_reader().read_to_end(&mut buf)?;
+
+        #[allow(clippy::cast_precision_loss)]
+        let dl_kib = buf.len() as f64 / 1024.0;
+        if dl_kib < 1024.0 {
+            info_end!("{:.02} KiB", Paint::new(dl_kib).fg(Green).bold());
+        } else {
+            info_end!("{:.02} MiB", Paint::new(dl_kib / 1024.0).fg(Green).bold());
+        }
+
+        Ok(buf)
+    }
+
+    /// Download tldr pages archives for directories that are out of date and update the checksum file.
+    fn download_and_verify(
+        &self,
+        mirror: &str,
+        languages: &[String],
+    ) -> Result<BTreeMap<String, PagesArchive>> {
+        let agent = ureq::builder()
+            .user_agent(USER_AGENT)
+            .try_proxy_from_env(true)
+            .build();
+
+        let sums = Self::get_asset(&agent, &format!("{mirror}/tldr.sha256sums"))?;
+        let sums_str = String::from_utf8_lossy(&sums);
+        let sum_map = Self::parse_sumfile(&sums_str)?;
+
+        let old_sumfile_path = self.dir.join("tldr.sha256sums");
+        let old_sums = fs::read_to_string(&old_sumfile_path).unwrap_or_default();
+        let old_sum_map = Self::parse_sumfile(&old_sums).unwrap_or_default();
+
         let mut langdir_archive_map = BTreeMap::new();
 
-        infoln!("downloading 'tldr.sha256sums'...");
-        let sums = agent.get(CHECKSUMS).call()?.into_string()?;
-        let lang_sum_map = Self::parse_sumfile(&sums)?;
-
         for lang in languages {
-            let sum = lang_sum_map.get(lang);
-            // Skip nonexistent languages.
-            if sum.is_none() {
+            let lang = &**lang;
+            let Some(sum) = sum_map.get(lang) else {
+                // Skip nonexistent languages.
+                continue;
+            };
+
+            let lang_dir = format!("pages.{lang}");
+            if Some(sum) == old_sum_map.get(lang) && self.subdir_exists(&lang_dir) {
+                infoln!("'pages.{lang}' is up to date");
                 continue;
             }
-            let sum = sum.unwrap();
-
-            infoln!("downloading 'tldr-pages.{lang}.zip'...");
-
-            let mut archive = vec![];
-            agent
-                .get(&format!("{BASE_ARCHIVE_URL}/tldr-pages.{lang}.zip"))
-                .call()?
-                .into_reader()
-                .read_to_end(&mut archive)?;
 
+            let archive = Self::get_asset(&agent, &format!("{mirror}/tldr-pages.{lang}.zip"))?;
             info_start!("validating sha256sums...");
             let actual_sum = util::sha256_hexdigest(&archive);
 
@@ -80,30 +117,33 @@ impl<'a> Cache<'a> {
 
             info_end!(" {}", Paint::new("OK").fg(Green).bold());
 
-            langdir_archive_map.insert(
-                format!("pages.{lang}"),
-                ZipArchive::new(Cursor::new(archive))?,
-            );
+            langdir_archive_map.insert(lang_dir, ZipArchive::new(Cursor::new(archive))?);
         }
 
+        fs::create_dir_all(self.dir)?;
+        File::create(&old_sumfile_path)?.write_all(&sums)?;
+
         Ok(langdir_archive_map)
     }
 
-    fn parse_sumfile(s: &str) -> Result<HashMap<String, String>> {
-        let mut map = HashMap::new();
+    fn parse_sumfile(s: &str) -> Result<HashMap<&str, &str>> {
+        // Subtract 3, because 3 lines are skipped in the loop.
+        let mut map = HashMap::with_capacity(s.lines().count().saturating_sub(3));
 
         for l in s.lines() {
             // The file looks like this:
-            // sha256sum     path/to/tldr-pages.lang.zip
-            // sha256sum     path/to/tldr-pages.lang.zip
+            // sha256sum     tldr-pages.lang.zip
+            // sha256sum     tldr-pages.lang.zip
             // ...
 
             let mut spl = l.split_whitespace();
             let sum = spl.next().ok_or_else(Error::parse_sumfile)?;
             let path = spl.next().ok_or_else(Error::parse_sumfile)?;
 
-            // Skip other files, the full archive, and the old English archive
-            // (tldr-pages.en.zip is now available).
+            // Skip other files, the full archive, and the old English archive.
+            // This map is used to detect languages available to download.
+            // Not skipping index.json makes "json" a language.
+            // Not skipping archives without a language in the filename makes "zip" a language.
             if !path.ends_with("zip")
                 || path.ends_with("tldr.zip")
                 || path.ends_with("tldr-pages.zip")
@@ -112,7 +152,7 @@ impl<'a> Cache<'a> {
             }
 
             let lang = path.split('.').nth(1).ok_or_else(Error::parse_sumfile)?;
-            map.insert(lang.to_string(), sum.to_string());
+            map.insert(lang, sum);
         }
 
         Ok(map)
@@ -130,21 +170,26 @@ impl<'a> Cache<'a> {
         info_start!("extracting '{lang_dir}'...");
 
         let mut n_downloaded = 0;
-        let files: Vec<String> = archive.file_names().map(String::from).collect();
 
-        for f in &files {
-            // Skip directory entries and files that are not in a directory (we want only pages).
-            if f.ends_with('/') || !f.contains('/') {
+        for i in 0..archive.len() {
+            let mut page = archive.by_index(i).unwrap();
+            let fname = page.name();
+
+            // Skip files that are not in a directory (we want only pages).
+            if !fname.contains('/') {
                 continue;
             }
 
-            let path = self.0.join(lang_dir).join(f);
-            fs::create_dir_all(path.parent().unwrap())?;
+            let path = self.dir.join(lang_dir).join(fname);
+
+            if fname.ends_with('/') {
+                fs::create_dir_all(&path)?;
+                continue;
+            }
 
-            let mut page = archive.by_name(f).unwrap();
             let mut file = File::create(&path)?;
-
             io::copy(&mut page, &mut file)?;
+
             n_downloaded += 1;
         }
 
@@ -162,34 +207,38 @@ impl<'a> Cache<'a> {
     }
 
     /// Delete the old cache and replace it with a fresh copy.
-    pub fn update(&self, languages: &[String]) -> Result<()> {
-        // (language_dir, pages_number_before_update)
-        let mut dirs_npages = HashMap::new();
-        let mut all_downloaded = 0;
-        let mut all_new = 0;
-
-        let mut languages = languages.to_vec();
+    pub fn update(&self, mirror: &str, languages: &mut Vec<String>) -> Result<()> {
         // Sort to always download archives in alphabetical order.
         languages.sort_unstable();
         // The user can put duplicates in the config file.
         languages.dedup();
-        let lang_dirs = util::languages_to_langdirs(&languages);
 
-        for lang_dir in &lang_dirs {
-            // `list_all_vec` can fail when `pages.en` is empty, hence the default of 0.
-            let n = self.list_all_vec(lang_dir).map(|v| v.len()).unwrap_or(0);
-            dirs_npages.insert(lang_dir.to_string(), n);
+        let archives = self.download_and_verify(mirror, languages)?;
+
+        if archives.is_empty() {
+            infoln!(
+                "there is nothing to do. Run 'tldr --clean-cache' if you want to force an update."
+            );
+            return Ok(());
         }
 
-        let archives = Self::download_and_verify(&languages)?;
-        self.clean()?;
+        let mut all_downloaded = 0;
+        let mut all_new = 0;
 
         for (lang_dir, mut archive) in archives {
+            // `list_all_vec` can fail when `pages.en` is empty, hence the default of 0.
             #[allow(clippy::cast_possible_truncation, clippy::cast_possible_wrap)]
+            let n_existing = self.list_all_vec(&lang_dir).map(|v| v.len()).unwrap_or(0) as i32;
+
+            let lang_dir_full = self.dir.join(&lang_dir);
+            if lang_dir_full.is_dir() {
+                fs::remove_dir_all(&lang_dir_full)?;
+            }
+
             self.extract_lang_archive(
                 &lang_dir,
                 &mut archive,
-                (*dirs_npages.get(&lang_dir).unwrap_or(&0)) as i32,
+                n_existing,
                 &mut all_downloaded,
                 &mut all_new,
             )?;
@@ -206,56 +255,56 @@ impl<'a> Cache<'a> {
 
     /// Delete the cache directory.
     pub fn clean(&self) -> Result<()> {
-        if !self.0.is_dir() {
+        if !self.dir.is_dir() {
             infoln!("cache does not exist, not cleaning.");
-            fs::create_dir_all(self.0)?;
+            fs::create_dir_all(self.dir)?;
             return Ok(());
         }
 
         infoln!("cleaning the cache directory...");
-        fs::remove_dir_all(self.0)?;
-        fs::create_dir_all(self.0)?;
+        fs::remove_dir_all(self.dir)?;
+        fs::create_dir_all(self.dir)?;
 
         Ok(())
     }
 
     /// Find out what platforms are available.
-    fn get_platforms(&self) -> Result<Vec<OsString>> {
-        let mut result = vec![];
+    fn get_platforms(&self) -> Result<&[OsString]> {
+        self.platforms
+            .get_or_try_init(|| {
+                let mut result = vec![];
 
-        for entry in fs::read_dir(self.0.join(ENGLISH_DIR))? {
-            let entry = entry?;
-            let path = entry.path();
-            let platform = path.file_name().unwrap();
+                for entry in fs::read_dir(self.dir.join(ENGLISH_DIR))? {
+                    let entry = entry?;
+                    let path = entry.path();
+                    let platform = path.file_name().unwrap();
 
-            result.push(platform.to_os_string());
-        }
+                    result.push(platform.to_os_string());
+                }
 
-        if result.is_empty() {
-            Err(Error::new(
-                "'pages.en' contains no platform directories. Please run 'tldr --update'.",
-            ))
-        } else {
-            // read_dir() order can differ across runs, so it's
-            // better to sort the Vec for consistency.
-            result.sort_unstable();
-            Ok(result)
-        }
+                if result.is_empty() {
+                    Err(Error::new(
+                        "'pages.en' contains no platform directories. Please run 'tldr --update'.",
+                    ))
+                } else {
+                    // read_dir() order can differ across runs, so it's
+                    // better to sort the Vec for consistency.
+                    result.sort_unstable();
+                    Ok(result)
+                }
+            })
+            .map(Vec::as_slice)
     }
 
     /// Find out what platforms are available and check if the provided platform exists.
-    fn get_platforms_and_check(&self, platform: &str) -> Result<Vec<OsString>> {
+    fn get_platforms_and_check(&self, platform: &str) -> Result<&[OsString]> {
         let platforms = self.get_platforms()?;
 
         if platforms.iter().all(|x| x != platform) {
             Err(Error::new(format!(
                 "platform '{platform}' does not exist.\n{} {}.",
                 Paint::new("Possible values:").bold(),
-                platforms
-                    .iter()
-                    .map(|x| x.to_string_lossy())
-                    .collect::<Vec<Cow<str>>>()
-                    .join(", ")
+                platforms.join(", ".as_ref()).to_string_lossy()
             )))
         } else {
             Ok(platforms)
@@ -268,7 +317,7 @@ impl<'a> Cache<'a> {
         P: AsRef<Path>,
     {
         for lang_dir in lang_dirs {
-            let path = self.0.join(lang_dir).join(&platform).join(fname);
+            let path = self.dir.join(lang_dir).join(&platform).join(fname);
 
             if path.is_file() {
                 return Some(path);
@@ -279,21 +328,16 @@ impl<'a> Cache<'a> {
     }
 
     /// Find all pages with the given name.
-    pub fn find(
-        &self,
-        name: &str,
-        languages: &mut Vec<String>,
-        platform: &str,
-    ) -> Result<Vec<PathBuf>> {
+    pub fn find(&self, name: &str, languages: &[String], platform: &str) -> Result<Vec<PathBuf>> {
         // https://github.com/tldr-pages/tldr/blob/main/CLIENT-SPECIFICATION.md#page-resolution
 
         let platforms = self.get_platforms_and_check(platform)?;
         let file = format!("{name}.md");
 
-        // We can't sort here - order is defined by the user.
-        languages.dedup_nosort();
-        let lang_dirs = util::languages_to_langdirs(languages);
         let mut result = vec![];
+        let mut lang_dirs: Vec<String> = languages.iter().map(|x| format!("pages.{x}")).collect();
+        // We can't sort here - order is defined by the user.
+        lang_dirs.dedup_nosort();
 
         // `common` is always searched, so we skip the search for the specified platform
         // if the user has requested only `common` (to prevent searching twice)
@@ -315,7 +359,7 @@ impl<'a> Cache<'a> {
                 continue;
             }
 
-            if let Some(path) = self.find_page_for(&file, &alt_platform, &lang_dirs) {
+            if let Some(path) = self.find_page_for(&file, alt_platform, &lang_dirs) {
                 if result.is_empty() {
                     let alt_platform = alt_platform.to_string_lossy();
 
@@ -345,13 +389,15 @@ impl<'a> Cache<'a> {
         P: AsRef<Path>,
         Q: AsRef<Path>,
     {
-        if let Ok(entries) = fs::read_dir(self.0.join(lang_dir.as_ref()).join(platform)) {
-            let entries = entries.map(|res| res.map(|ent| ent.file_name()));
-            Ok(entries.collect::<io::Result<Vec<OsString>>>()?)
-        } else {
+        match fs::read_dir(self.dir.join(lang_dir.as_ref()).join(platform)) {
+            Ok(entries) => {
+                let entries = entries.map(|res| res.map(|ent| ent.file_name()));
+                Ok(entries.collect::<io::Result<Vec<OsString>>>()?)
+            }
             // If the directory does not exist, return an empty Vec instead of an error
             // (some platform directories do not exist in some translations).
-            Ok(vec![])
+            Err(e) if e.kind() == io::ErrorKind::NotFound => Ok(vec![]),
+            Err(e) => Err(e.into()),
         }
     }
 
@@ -375,7 +421,8 @@ impl<'a> Cache<'a> {
     }
 
     /// List all pages in English for `platform` and common.
-    pub fn list_platform(&self, platform: &str) -> Result<()> {
+    pub fn list_for(&self, platform: &str) -> Result<()> {
+        // This is here just to check if the platform exists.
         self.get_platforms_and_check(platform)?;
 
         let pages = if platform == "common" {
@@ -398,7 +445,7 @@ impl<'a> Cache<'a> {
         let mut result = vec![];
 
         for platform in self.get_platforms()? {
-            result.append(&mut self.list_dir(&platform, &lang_dir)?);
+            result.append(&mut self.list_dir(platform, &lang_dir)?);
         }
 
         Ok(result)
@@ -409,19 +456,47 @@ impl<'a> Cache<'a> {
         Self::print_basenames(self.list_all_vec(ENGLISH_DIR)?)
     }
 
+    /// List platforms (used in shell completions).
+    pub fn list_platforms(&self) -> Result<()> {
+        let platforms = self.get_platforms()?.join("\n".as_ref());
+        writeln!(io::stdout(), "{}", platforms.to_string_lossy())?;
+        Ok(())
+    }
+
+    /// List languages (used in shell completions).
+    pub fn list_languages(&self) -> Result<()> {
+        let languages = fs::read_dir(self.dir)?
+            .filter(|res| res.is_ok() && res.as_ref().unwrap().path().is_dir())
+            .map(|res| res.unwrap().file_name());
+        let mut stdout = io::stdout().lock();
+
+        for lang in languages {
+            let lang = lang.to_string_lossy();
+            let lang = lang.strip_prefix("pages.").unwrap_or(&lang);
+
+            writeln!(stdout, "{lang}")?;
+        }
+
+        Ok(())
+    }
+
     /// Show cache information.
     pub fn info(&self, cfg: &Config) -> Result<()> {
         let mut n_map = BTreeMap::new();
         let mut n_total = 0;
 
-        for lang_dir in fs::read_dir(self.0)? {
-            let lang_dir = lang_dir?.file_name();
+        for lang_dir in fs::read_dir(self.dir)? {
+            let lang_dir = lang_dir?;
+            if !lang_dir.path().is_dir() {
+                continue;
+            }
+            let lang_dir = lang_dir.file_name();
             let n = self.list_all_vec(&lang_dir)?.len();
 
-            let str = lang_dir.to_string_lossy();
-            let str = str.split_once('.').unwrap_or(("", &str)).1;
+            let lang = lang_dir.to_string_lossy();
+            let lang = lang.strip_prefix("pages.").unwrap_or(&lang);
 
-            n_map.insert(str.to_string(), n);
+            n_map.insert(lang.to_string(), n);
             n_total += n;
         }
 
@@ -431,7 +506,7 @@ impl<'a> Cache<'a> {
         writeln!(
             stdout,
             "Cache: {} (last update: {} ago)",
-            Paint::new(self.0.display()).fg(Red),
+            Paint::new(self.dir.display()).fg(Red),
             Paint::new(util::duration_fmt(age)).fg(Green).bold()
         )?;
 
@@ -469,12 +544,24 @@ impl<'a> Cache<'a> {
 
     /// Get the age of the cache.
     pub fn age(&self) -> Result<Duration> {
-        fs::metadata(self.0)?.modified()?.elapsed().map_err(|_| {
-            Error::new(
-                "the system clock is not functioning correctly.\n\
-                Modification time of the cache is later than the current system time.\n\
-                Please fix your system clock.",
-            )
-        })
+        self.age
+            .get_or_try_init(|| {
+                let sumfile = self.dir.join("tldr.sha256sums");
+                let metadata = if sumfile.is_file() {
+                    fs::metadata(&sumfile)
+                } else {
+                    // The sumfile is not available, fall back to the base directory.
+                    fs::metadata(self.dir)
+                }?;
+
+                metadata.modified()?.elapsed().map_err(|_| {
+                    Error::new(
+                        "the system clock is not functioning correctly.\n\
+                        Modification time of the cache is later than the current system time.\n\
+                        Please fix your system clock.",
+                    )
+                })
+            })
+            .copied()
     }
 }
